/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package boswarsFontes;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Scanner;

/**
 *
 * @author Rodrigo
 */
public class Individual implements Comparable<Individual>, Cloneable {

    /**
     * Attributes
     */
    
    private double fitness;
    private double sFitness;
    private int nBuild;
    private int nKill;
    private int nRaz;
    private int nUnit;
    private int score;
    private int cycle;
    private int vFac;
    private int camp;
    private int aFac;
    private int hosp;
    private int id;
    private int pos;
    private int reward;
    private String name;
    private int nMatches;
    private int nVic;
    
    /**
     * Chromosome of actions outside the loop
     */
    public ArrayList<Gene> chromoOut;
    
    /**
     * Chromosome of actions inside the loop
     */
    public ArrayList<Gene> chromoIn;

    /**
     * Constructor - Do nothing
     */
    public Individual() {
    }

    /**
     * Constructor - Initialize individual
     *
     * @param id Identification of the individual
     * @param pos Position of the individual on the map
     * @para decode True if it is to be decoded, false otherwise
     */
    public Individual(int id, int pos, boolean decode) throws FileNotFoundException {
        this.setId(id);
        this.setPos(pos);
        this.resetCounters();

        chromoOut = new ArrayList();
        chromoIn = new ArrayList();

        this.initializeChromoRandomSize(decode);
    }

    /**
     * Clone an Individual
     *
     * @return This individual
     * @throws CloneNotSupportedException
     */
    @Override
    protected Object clone() throws CloneNotSupportedException {
        Individual c = (Individual) super.clone();

        c.chromoIn = new ArrayList();
        c.chromoOut = new ArrayList();
        /**
         * Copy chromosomes
         */
        for (Gene g : chromoIn) {
            c.chromoIn.add((Gene) g.clone());
        }

        for (Gene g : chromoOut) {
            c.chromoOut.add((Gene) g.clone());
        }

        return c;
    }

    /**
     * Initialize individual
     */
    public void initialize(boolean decode) {
    }

    /**
     * Initialize individual with chromosome with random length
     */
    public final void initializeChromoRandomSize(boolean decode) throws FileNotFoundException {
        int sizeOut, sizeIn;

        /**
         * Raffle size of the chromosomes
         */
        sizeOut = Util.rmd.nextInt(61) + 20;
        sizeIn = Util.rmd.nextInt(16) + 5;

        /**
         * Create Chromosomes
         */
        for (int i = 0; i < sizeOut; i++) {
            this.chromoOut.add(new Gene());
            this.chromoOut.get(i).raffle(false, false);
        }
        for (int i = 0; i < sizeIn; i++) {
            this.chromoIn.add(new Gene());
            this.chromoIn.get(i).raffle(true, false);
        }

        this.updateName();

        if (decode == true) {
            this.decode();
        }
    }

    /**
     * Evaluate individual - Call the game, recover statistics, generate fitness
     */
    public void evaluate(Individual enemy) throws FileNotFoundException {
        this.nMatches++;
        this.recoverStatistics();
        this.generateFitness(enemy);
        
         if(this.getReward() == 1){
             nVic+=1;
         }
    }

    /**
     * Recover statistics data of a match
     */
    public void recoverStatistics() throws FileNotFoundException {
        String file = "indPos" + this.getPos() + ".bwdb";

        Scanner sc = new Scanner(new File(DataHandler.object().dir + file));
        this.setnBuild(sc.nextInt());
        this.setnKill(sc.nextInt());
        this.setnRaz(sc.nextInt());
        this.setnUnit(sc.nextInt());
        this.setScore(sc.nextInt());
        this.setReward(sc.nextInt());
        this.setCycle(sc.nextInt());
        sc.close();

    }

    /**
     * Update the name of the individual
     */
    public void updateName() {
        this.name = "pp" + this.getPos() + "ind" + this.getId();
    }

    /**
     * Generate fitness value
     */
    public void generateFitness(Individual enemy) {
        double f;
        
        //TODO funcao fitness
        if (this.getReward() == 1) {
            f = 1 - (double) this.getCycle() / (double) (DataHandler.MAXCYCLES * 2);
        }
        else{
            f = (double) this.getCycle() / (double) (DataHandler.MAXCYCLES * 2);
        }
        
        fitness = f;
    }

    /**
     * Prepare individual for a single match It is just decode and register it
     */
    public void prepareForBattle() throws FileNotFoundException {
        
        this.updateName();
        this.decode();
        this.register();
        
        if(this.getReward() == 0){
            this.resetCounters();
        }
    }

    /**
     * Clear chromosomes and attributes
     */
    public void clear() {
        this.chromoIn.clear();
        this.chromoOut.clear();
    }

    /**
     * Reset counters of the individual
     */
    public final void resetCounters() {
        this.setCycle(0);
        this.setFitness(0);
        this.setsFitness(0);
        this.setScore(0);
        this.setReward(0);
        this.setnBuild(0);
        this.setnKill(0);
        this.setnRaz(0);
        this.setnUnit(0);
        this.setnMatches(0);
        this.setnVic(0);
    }

    /**
     * Crossover operator
     *
     * @param Ind Parent to make the crossover
     * @return New individual created
     */
    public void crossover(Individual ind1, Individual ind2) throws CloneNotSupportedException {

        /**
         * Cross chromoIn
         */
        if (ind1.chromoIn.size() >= ind2.chromoIn.size()) {
            cross(this.chromoIn, ind1.chromoIn, ind2.chromoIn);
        } else {
            cross(this.chromoIn, ind2.chromoIn, ind1.chromoIn);
        }

        /**
         * Cross chromoOut
         */
        if (ind1.chromoOut.size() >= ind2.chromoOut.size()) {
            cross(this.chromoOut, ind1.chromoOut, ind2.chromoOut);
        } else {
            cross(this.chromoOut, ind2.chromoOut, ind1.chromoOut);
        }


    }

    public void cross(ArrayList<Gene> child, ArrayList<Gene> bigger, ArrayList<Gene> other) throws CloneNotSupportedException {

        /**
         * Clear child and perform crossover
         */
        child.clear();
        //TODO Insira o(s) metodo(s) de crossover aqui!
        //Nao esqueçam de clonar o gene do pai!
    }

    /**
     * Mutation operator
     *
     */
    public void mutation() throws FileNotFoundException {
        //50% of chance to perform mutation
        if (Util.rmd.nextInt(100) < 50) {
            performMutation(this.chromoOut, false);
            performMutation(this.chromoIn, true);
        }
        
    }

    protected void performMutation(ArrayList<Gene> chromo, boolean loop) throws FileNotFoundException {
        
        //TODO Insira o método de Mutacao aqui
    
    }

    /**
     * Fix Chromosome
     */
    public void fixChromosome() {
        fixChromoOut();
      //  fixChromoIn();
    }

    protected void fixChromoOut() {
        int size = (int) ((double) chromoOut.size() / 3);

       // fix(0, size);
        fix(size, size * 2);
        fix(size, chromoOut.size());

    }

    protected void fix(int ini, int end) {
        boolean fvFac = false;
        boolean faFac = false;
        boolean fcamp = false;
        boolean flag = false;
        Gene ng;

        for (int i = ini; i < end; i++) {
            if (chromoOut.get(i).getType() == DataHandler.CREATE_BUILDING) {
                switch (chromoOut.get(i).getBuilding()) {
                    case "unit-camp":
                        fcamp = true;
                        break;
                    case "unit-aircraftfactory":
                        faFac = true;
                        break;
                    case "unit-vfac":
                        fvFac = true;
                        break;
                }
            }
        }

        if (fvFac == true && faFac == true && fcamp == true) {
            return;
        }

        for (int i = ini; i < end; i++) {
            if (chromoOut.get(i).getType() == DataHandler.CREATE_BATTALION) {
                if (chromoOut.get(i).getBuilding().equals("unit-camp") && fcamp == false) {
                    fcamp = true;
                    flag = true;
                } else if (chromoOut.get(i).getBuilding().equals("unit-aircraftfactory") && faFac == false) {
                    faFac = true;
                    flag = true;
                } else if (chromoOut.get(i).getBuilding().equals("unit-vfac") && fvFac == false) {
                    fvFac = true;
                    flag = true;
                }

                if (flag == true) {
                    flag = false;
                    ng = new Gene();
                    ng.setType(DataHandler.CREATE_BUILDING);
                    ng.setBuilding(chromoOut.get(i).getBuilding());
                    if (i == ini) {
                        chromoOut.add(ini, ng);
                    } else {
                        chromoOut.add(Util.rmd.nextInt(i - ini) + ini, ng);
                    }

                }
            }
        }
    }

    protected void fixChromoIn() {
        boolean fvFacOut = false;
        boolean faFacOut = false;
        boolean fcampOut = false;

        boolean fvFacIn = false;
        boolean faFacIn = false;
        boolean fcampIn = false;

        Gene ng;

        for (Gene g : chromoOut) {
            if (g.getType() == DataHandler.CREATE_BUILDING) {
                if ((g.getBuilding().equals("unit-aircraftfactory"))) {
                    faFacOut = true;
                }

                if ((g.getBuilding().equals("unit-camp"))) {
                    fcampOut = true;
                }

                if ((g.getBuilding().equals("unit-vfac"))) {
                    fvFacOut = true;
                }
            }
        }

        for (Gene g : chromoIn) {
            if ((g.getBuilding().equals("unit-aircraftfactory"))) {
                faFacIn = true;
            }

            if ((g.getBuilding().equals("unit-camp"))) {
                fcampIn = true;
            }

            if ((g.getBuilding().equals("unit-vfac"))) {
                fvFacIn = true;
            }
        }

        if (faFacOut == false && faFacIn == true) {
            ng = new Gene();
            ng.setBuilding("unit-aircraftfactory");
            chromoOut.add(Util.rmd.nextInt(chromoOut.size()), ng);
        }

        if (fvFacOut == false && fvFacIn == true) {
            ng = new Gene();
            ng.setBuilding("unit-vfac");
            chromoOut.add(Util.rmd.nextInt(chromoOut.size()), ng);
        }

        if (fcampOut == false && fcampIn == true) {
            ng = new Gene();
            ng.setBuilding("unit-camp");
            chromoOut.add(Util.rmd.nextInt(chromoOut.size()), ng);
        }
    }

    /**
     * Register individual in a match
     */
    public void register() throws FileNotFoundException {
        ArrayList<String> newFile = new ArrayList();
        File setup = new File(DataHandler.object().dir + "maps\\" + DataHandler.object().mapName + ".map\\setup.sms");
        int contToken = -1;
        int index = 0;
        String token = "--*";

        //Read File
        Scanner sc = new Scanner(setup);
        while (sc.hasNextLine()) {
            newFile.add(sc.nextLine());
        }
        sc.close();

        //Delete
        setup.delete();

        //Insert new lines
        while (true) {
            if (newFile.get(index).equals(token)) {
                contToken++;
            }

            //position found
            if (contToken == this.getPos()) {
                newFile.set(index + 1, "Players[" + this.getPos() + "].AiName = \"ai-" + this.getName() + "\"");
                break;
            }
            index++;
        }

        //Create new file
        setup = new File(DataHandler.object().dir + "maps\\" + DataHandler.object().mapName + ".map\\setup.sms");
        PrintStream out = new PrintStream(new FileOutputStream(setup, false));

        for (String line : newFile) {
            out.println(line);
        }
        out.close();
    }

    /**
     * Decode individual into a LUA script
     */
    public void decode() throws FileNotFoundException {
        
        this.setCamp(0);
        this.setaFac(0);
        this.setHosp(0);
        this.setvFac(0);
        
        String dir = DataHandler.object().dir + "scripts\\ais\\";
        PrintStream out = new PrintStream(new FileOutputStream(new File(dir + this.getName() + ".lua"), false));


        out.printf("local player \nlocal ai_pos \nlocal ai_loop_pos\n");

        /*AiLoop*/
        out.printf("local function AiLoop(loop_funcs, loop_pos) \n");
        out.printf("local ret\n player = AiPlayer() + 1\n");
        out.printf("while (true) do \nret = loop_funcs[loop_pos[player]]()\n");
        out.printf("if (ret) then \nbreak \nend \nloop_pos[player] = loop_pos[player] + 1\n");
        out.printf("end\n return true\n end\n\n");

        /*InitAiScripts*/
        out.printf("function InitAiScripts_%s()\n", this.name);
        out.printf("ai_pos = {1, 1, 1, 1, 1, 1, 1, 1}\n");
        out.printf("ai_loop_pos = {1, 1, 1, 1, 1, 1, 1, 1} \nhotspotexists = nil \nend\n\n");

        /*Script Looping*/
        out.printf("local ai_loop_funcs = {\n");

        /*Decode Script*/
        for (int i = 0; i < this.chromoIn.size(); i++) {
            if ((i > 0) && (chromoIn.get(i - 1).isHE() == false)
                    && (chromoIn.get(i).getForceCode() != chromoIn.get(i - 1).getForceCode())) {
                out.printf("function() return AiWaitForce(%d) end,\n", this.chromoIn.get(i - 1).getForceCode());
                out.printf("function() return AiAttackWithForce(%d) end,\n", this.chromoIn.get(i - 1).getForceCode());
            }

            if (chromoIn.get(i).isHE() == true) {
                out.printf("function() return AiNeed(\"%s\") end,\n", this.chromoIn.get(i).getUnit());
            } else {
                out.printf("function() return AiForce(%d,{\"%s\",%d}) end,\n", this.chromoIn.get(i).getForceCode(),
                        this.chromoIn.get(i).getUnit(),
                        this.chromoIn.get(i).getSizeUnit());

                //if it is the last gene
                if ((i == (chromoIn.size() - 1)) && (chromoIn.get(i).isHE() == false)) {
                    out.printf("function() return AiWaitForce(%d) end,\n", this.chromoIn.get(i).getForceCode());
                    out.printf("function() return AiAttackWithForce(%d) end,\n", this.chromoIn.get(i).getForceCode());
                }
            }
        }
        out.printf("function() \nai_loop_pos[player] = 0; \nreturn false \nend,\n}\n\n");

        /*HotSpotExist*/
        out.printf("local function HotSpotExists()\nif (hotspotexists == nil) then\n");
        out.printf("local hotspot = UnitTypeByIdent(\"unit-hotspot\")\n");
        out.printf("local count = Players[PlayerNumNeutral].UnitTypesCount[hotspot.Slot]\n");
        out.printf("hotspotexists = (count %c= 0)\nend \nreturn hotspotexists \nend\n\n", 126);

        /*CheckPower*/
        out.printf("local function CheckPower()\n");
        out.printf("local nuke = UnitTypeByIdent(\"unit-nukepowerplant\")\n");
        out.printf("local count = Players[AiPlayer()].UnitTypesCount[nuke.Slot]\n");
        out.printf("if (count %c= 0) then \nreturn nil \nend \nreturn \"unit-powerplant\" \nend\n\n", 126);

        /*GetBuildOrder*/
        out.printf("local function GetBuildOrder() \nlocal order = {}\n");
        out.printf("if (not HotSpotExists()) then \norder[1] = CheckPower()\n order[2] = nil\n");
        out.printf("elseif (Players[AiPlayer()].MagmaStored < 300) then\n");
        out.printf("order[1] = \"unit-magmapump\" \norder[2] = CheckPower() \nelse\n");
        out.printf("order[1] = CheckPower()\norder[2] = \"unit-magmapump\" \nend \nreturn order \n end\n\n");

        /*Script*/
        out.printf("local ai_funcs = { \nfunction() AiDebug(false) return false end,\n");
        out.printf(" -- Build magma pump or power plant first depending on resources\n");
        out.printf("function() \norder = GetBuildOrder() \nif (order[1] %c= nil) then\n", 126);
        out.printf("return AiNeed(order[1]) \nelse \nreturn false \nend \nend, \n\n");

        out.printf("function() \norder = GetBuildOrder() \nif (order[1] %c= nil) then\n", 126);
        out.printf("return AiWait(order[1]) \nelse \nreturn false \nend \nend, \n\n");

        out.printf("function() \norder = GetBuildOrder() \nif (order[2] %c= nil) then\n", 126);
        out.printf("return AiNeed(order[2]) \nelse \nreturn false \nend \nend, \n\n");

        out.printf("function() \norder = GetBuildOrder() \nif (order[2] %c= nil) then\n", 126);
        out.printf("return AiWait(order[2]) \nelse \nreturn false \nend \nend,\n\n");

        out.printf("function() return AiSet(\"unit-engineer\", 6) end,\n");
        out.printf("function() return AiNeed(\"unit-vault\") end,\n");
        out.printf("function() return AiNeed(\"unit-magmapump\") end,\n");
        out.printf("function() return AiWait(\"unit-vault\") end,\n\n");

        /*Decode Script*/
        for (int i = 0; i < this.chromoOut.size(); i++) {


            if ((i > 0) && (chromoOut.get(i - 1).getType() == DataHandler.CREATE_BATTALION)
                    && (chromoOut.get(i - 1).isHE() == false)
                    && (chromoOut.get(i).getForceCode() != chromoOut.get(i - 1).getForceCode())) {
                out.printf("function() return AiWaitForce(%d) end,\n", this.chromoOut.get(i - 1).getForceCode());
                out.printf("function() return AiAttackWithForce(%d) end ,\n", this.chromoOut.get(i - 1).getForceCode());
            }

            if (chromoOut.get(i).getType() == DataHandler.CREATE_BUILDING) {
                out.printf("function() return AiNeed(\"%s\") end,\n", chromoOut.get(i).getBuilding());
            } else if (chromoOut.get(i).getType() == DataHandler.CREATE_BATTALION) {
                    //verify it a building is needed

                 if (chromoOut.get(i).getBuilding().equals("unit-camp") && (this.getCamp() == 0)
                         
                            || (chromoOut.get(i).getBuilding().equals("unit-vfac") && (this.getvFac() == 0))
                            || (chromoOut.get(i).getBuilding().equals("unit-aircraftfactory") && (this.getaFac() == 0))
                      )
                          {
                        out.printf("function() return AiNeed(\"%s\") end,\n", this.chromoOut.get(i).getBuilding());
                }
                 
                
                if (chromoOut.get(i).isHE()) {
                    out.printf("function() return AiNeed(\"%s\") end,\n", this.chromoOut.get(i).getUnit());
                } else {

                    out.printf("function() return AiForce(%d,{\"%s\",%d}) end,\n", this.chromoOut.get(i).getForceCode(),
                            this.chromoOut.get(i).getUnit(),
                            this.chromoOut.get(i).getSizeUnit());

                    //if it is the last gene
                    if (i == (chromoOut.size() - 1)) {
                        out.printf("function() return AiWaitForce(%d) end,\n", this.chromoOut.get(i).getForceCode());
                        out.printf("function() return AiAttackWithForce(%d) end ,\n", this.chromoOut.get(i).getForceCode());
                    }
                }
            }

            //update flags of buildings
            switch (this.chromoOut.get(i).getBuilding()) {
                case "unit-camp":
                    this.setCamp(1);
                    break;
                case "unit-vfac":
                    this.setvFac(1);
                    break;
                case "unit-aircraftfactory":
                    this.setaFac(1);
                    break;
                case "unit-hosp":
                    this.setHosp(1);
                    break;
            }

        }

        out.printf("function() return AiLoop(ai_loop_funcs, ai_loop_pos) end, \n}\n\n");

        /*Constructor and Register AI*/
        out.printf("function Ai%s() \nreturn AiLoop(ai_funcs, ai_pos) \nend\n\n", this.getName());
        out.printf("RegisterAi(\"%s\",\"%s\",Ai%s,InitAiScripts_%s)\n", this.getName(), this.getName(), this.getName(), this.getName());

        out.close();
    }

    /**
     * Compare to individuals
     */
    @Override
    public int compareTo(Individual ind) {
        if (ind.getFitness() > this.getFitness()) {
            return 1;
        } else if (ind.getFitness() < this.getFitness()) {
            return -1;
        } else {
            return 0;
        }
    }

    /**
     * @return the fitness
     */
    public double getFitness() {
        return fitness;
    }

    /**
     * @param fitness the fitness to set
     */
    public final void setFitness(double fitness) {
        this.fitness = fitness;
    }

    /**
     * @return the nBuild
     */
    public int getnBuild() {
        return nBuild;
    }

    /**
     * @param nBuild the nBuild to set
     */
    public final void setnBuild(int nBuild) {
        this.nBuild = nBuild;
    }

    /**
     * @return the nKill
     */
    public int getnKill() {
        return nKill;
    }

    /**
     * @param nKill the nKill to set
     */
    public final void setnKill(int nKill) {
        this.nKill = nKill;
    }

    /**
     * @return the nRaz
     */
    public int getnRaz() {
        return nRaz;
    }

    /**
     * @param nRaz the nRaz to set
     */
    public final void setnRaz(int nRaz) {
        this.nRaz = nRaz;
    }

    /**
     * @return the nUnit
     */
    public int getnUnit() {
        return nUnit;
    }

    /**
     * @param nUnit the nUnit to set
     */
    public final void setnUnit(int nUnit) {
        this.nUnit = nUnit;
    }

    /**
     * @return the score
     */
    public int getScore() {
        return score;
    }

    /**
     * @param score the score to set
     */
    public final void setScore(int score) {
        this.score = score;
    }

    /**
     * @return the cycle
     */
    public int getCycle() {
        return cycle;
    }

    /**
     * @param cycle the cycle to set
     */
    public final void setCycle(int cycle) {
        this.cycle = cycle;
    }

    /**
     * @return the reward
     */
    public int getReward() {
        return reward;
    }

    /**
     * @param reward the reward to set
     */
    public final void setReward(int reward) {
        this.reward = reward;
    }

    /**
     * @return the vFac
     */
    public int getvFac() {
        return vFac;
    }

    /**
     * @param vFac the vFac to set
     */
    public void setvFac(int vFac) {
        this.vFac = vFac;
    }

    /**
     * @return the camp
     */
    public int getCamp() {
        return camp;
    }

    /**
     * @param camp the camp to set
     */
    public final void setCamp(int camp) {
        this.camp = camp;
    }

    /**
     * @return the aFac
     */
    public int getaFac() {
        return aFac;
    }

    /**
     * @param aFac the aFac to set
     */
    public void setaFac(int aFac) {
        this.aFac = aFac;
    }

    /**
     * @return the hosp
     */
    public int getHosp() {
        return hosp;
    }

    /**
     * @param hosp the hosp to set
     */
    public void setHosp(int hosp) {
        this.hosp = hosp;
    }

    /**
     * @return the id
     */
    public int getId() {
        return id;
    }

    /**
     * @param id the id to set
     */
    public final void setId(int id) {
        this.id = id;
    }

    /**
     * @return the pos
     */
    public int getPos() {
        return pos;
    }

    /**
     * @param pos the pos to set
     */
    public final void setPos(int pos) {
        this.pos = pos;
    }

    /**
     * @return the name
     */
    public String getName() {
        return name;
    }

    /**
     * @param name the name to set
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @return the nMatches
     */
    public int getnMatches() {
        return nMatches;
    }

    /**
     * @param nMatches the nMatches to set
     */
    public void setnMatches(int nMatches) {
        this.nMatches = nMatches;
    }

    /**
     * @return the sFitness
     */
    public double getsFitness() {
        return sFitness;
    }

    /**
     * @param sFitness the sFitness to set
     */
    public void setsFitness(double sFitness) {
        this.sFitness = sFitness;
    }

    /**
     * @return the nVic
     */
    public int getnVic() {
        return nVic;
    }

    /**
     * @param nVic the nVic to set
     */
    public void setnVic(int nVic) {
        this.nVic = nVic;
    }
}
