/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package boswarsFontes;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;

/**
 *
 * @author Rodrigo
 */
public class Population {

    /**
     * List of individuals
     */
    public ArrayList<Individual> player = null;
    
    private int id;
    private int size;
    private int cni;
    private String bpFileName;
    private String fvFileName;
    private int iP1;
    private int iP2;
    private int currPlayer;
    private int qPlayers;
    private int fVic;

    /**
     * Constructor - Do nothing.
     */
    public Population() {
    }

    /**
     * Constructor
     *
     * @param id Indicade position of individuals on a match
     * @see Individual
     * @param size Size of the population
     * @param decode True to decode individuals immediately, false otherwise
     */
    public Population(int id, int size, boolean decode) throws FileNotFoundException {
        /**
         * Set initial value to the attributes.
         */
        this.setId(id);
        this.setSize(size);
        this.setiP1(0);
        this.setiP2(0);
        this.setqPlayers(4);
        this.setfVic(0);
        this.setCni(7);
        this.setFvFileName("fvLogPop" + this.id + ".csv");

        player = new ArrayList<>();


        this.initialize(decode);
    }

    /**
     * Initialize population.
     */
    public void initialize(boolean decode) throws FileNotFoundException {
        /**
         * Create Individuals
         */
        
        for (int i = 0; i < this.getSize(); i++) {
            this.player.add(new Individual(i, this.getId(), decode));
        }

        this.resetCounters();
    }

    /**
     * Evolve population
     *
     * @param counterMatch Number of the match to be played by the new
     * individual
     * @return new individual created
     */
    public void evolve(int counterMatch) throws CloneNotSupportedException, FileNotFoundException {
        int c = counterMatch % this.getqPlayers();
        
        this.currPlayer  = (this.getSize() - 4 + c);
        
        //Metodo para selecionar os pais
        this.roulette();
        
        if(c == 0){
            this.player.get(currPlayer).crossover(player.get(0), player.get(iP1));
        }else{
            this.player.get(currPlayer).crossover(player.get(iP1), player.get(iP2));
        }
        
        this.player.get(currPlayer).mutation();
        this.player.get(currPlayer).fixChromosome();

    }

    /**
     * Update population.
     */
    public void update(int counterMatch) {

        /**
         * True of it is the first battle won by 'pop', false otherwise
         */
        if ((this.getfVic() == 0) && (this.player.get(getCurrPlayer()).getReward() != 0)) {
            this.setfVic(counterMatch + 1);
        }

        /**
         * Order population if necessary
         */
        if (counterMatch % this.getqPlayers() == (getqPlayers() -1)) {
            this.sort();
        }
    }

    /**
     * Clear population.
     */
    public void clear() {
        this.player.clear();
        this.player = new ArrayList();
    }

    /**
     * Perform roulette selection.
     */
    public void roulette() {

        //Implemente o método para selecionar o indice dos pais
        this.iP1 = 0;
        this.iP2 = 0;
        
        /*this.iP1 = -1;
        this.iP2 = -1;*/
    }

    
    /**
     * Find best individuals between two population.
     *
     * @param p1 Population #1 to be analyzed
     * @param p2 Population #2 to be analyzed
     */
    public void findBest(Population p1, Population p2) throws CloneNotSupportedException {
        int i, j;

        /**
         * i - index of p1 j - index of p2
         */
        i = 0;
        j = 0;

        /**
         * Clear 'this.pop before copying'
         */
        this.clear();


        /**
         * Order p1 and p2 before making comparisons.
         */
        p1.sort();
        p2.sort();

        /**
         * Make comparisons and copy individuals.
         */
        for (int c = 0; c < this.getSize(); c++) {
            if (p1.player.get(i).getFitness() >= p2.player.get(j).getFitness()) {
                this.player.add((Individual) p1.player.get(i).clone());
                i++;
            } else {
                this.player.add((Individual) p2.player.get(j).clone());
                j++;
            }
            this.player.get(c).setPos(this.getId());
        }
        this.sort();
    }

    /**
     * Sort Individuals
     */
    public void sort(){
        Collections.sort(this.player);
        for(int i = 0; i < player.size();i++){
            player.get(i).setId(i);
            player.get(i).updateName();
        }
    }
    
    
    /**
     * Create First Victory log file.
     */
    public void createFVLogFile() throws FileNotFoundException {
        PrintStream out = new PrintStream(new FileOutputStream(new File(this.getFvFileName()),false));
        out.printf(";;;First Victory Counter\n\n");
        out.close();
    }

    /**
     * Update Fisrt Victory log file
     */
    public void updateFVLogFile() throws FileNotFoundException {
        PrintStream out = new PrintStream(new FileOutputStream(new File(this.getFvFileName()), true));
        out.printf(";;;%d\n", this.getfVic());
        out.close();
    }

    /**
     * Reset Counters of population
     */
    public void resetCounters() {
        this.setfVic(0);
    }

    /**
     * @return the id
     */
    public int getId() {
        return id;
    }

    /**
     * @param id the id to set
     */
    public final void setId(int id) {
        this.id = id;
    }

    /**
     * @return the size
     */
    public final int getSize() {
        return size;
    }

    /**
     * @param size the size to set
     */
    public final void setSize(int size) {
        this.size = size;
    }

    /**
     * @return the cni
     */
    public int getCni() {
        return cni;
    }

    /**
     * @param cni the cni to set
     */
    public final void setCni(int cni) {
        this.cni = cni;
    }

    /**
     * @return the bpFileName
     */
    public String getBpFileName() {
        return bpFileName;
    }

    /**
     * @param bpFileName the bpFileName to set
     */
    public void setBpFileName(String bpFileName) {
        this.bpFileName = bpFileName;
    }

    /**
     * @return the fvFileName
     */
    public String getFvFileName() {
        return fvFileName;
    }

    /**
     * @param fvFileName the fvFileName to set
     */
    public final void setFvFileName(String fvFileName) {
        this.fvFileName = fvFileName;
    }

    /**
     * @return the iP1
     */
    public int getiP1() {
        return iP1;
    }

    /**
     * @param iP1 the iP1 to set
     */
    public final void setiP1(int iP1) {
        this.iP1 = iP1;
    }

    /**
     * @return the iP2
     */
    public int getiP2() {
        return iP2;
    }

    /**
     * @param iP2 the iP2 to set
     */
    public final void setiP2(int iP2) {
        this.iP2 = iP2;
    }

    /**
     * @return the currPlayer
     */
    public int getCurrPlayer() {
        return currPlayer;
    }

    /**
     * @param currPlayer the currPlayer to set
     */
    public void setCurrPlayer(int currPlayer) {
        this.currPlayer = currPlayer;
    }

    /**
     * @return the qPlayers
     */
    public final int getqPlayers() {
        return qPlayers;
    }

    /**
     * @param qPlayers the qPlayers to set
     */
    public final void setqPlayers(int qPlayers) {
        this.qPlayers = qPlayers;
    }

    /**
     * @return the fVic
     */
    public int getfVic() {
        return fVic;
    }

    /**
     * @param fVic the fVic to set
     */
    public final void setfVic(int fVic) {
        this.fVic = fVic;
    }
}
